/*************************************************************************/
/* Reads a pdb file in the cns pdb format                                */
/* Calculates the theoretical PERE by numerical integration of the 1/r6  */
/* distance dependency. Calculates the distance to the closest para-     */
/* magnetic center.                                                      */
/*                                                                       */
/* Programmer:   Tobias Madl and Klaus Zangger                           */
/* Date:         2008                                                    */
/* Organization:           Institute of Chemistry                        */
/*                         Organic and Bioorganic Chemistry              */
/*                         Structural and Analytical Organic Chemistry   */
/*                         Heinrichstrasse 28                            */
/*                         A-8010 Graz                                   */
/*                         Austria                                       */
/*                         Europe                                        */
/*                                                                       */
/*************************************************************************/

#include <stdio.h>
#include <stddef.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>

int main() {

  unsigned _stklen = 4194304;  /* need a 1MB stack */
  int      na,i,j,k,q,natoms,c,posit,flag_change,
           atomnum[100000],resnum[100000];
  float    rpara,xmin,ymin,zmin,f,f_c,xmax,ymax,zmax,PREmaxcalc,PREmaxmeas,norm,
           *parapx,*parapy,*parapz,x,y,z,layer,spread,radius,d,
           paravol[100000],parainfl[100000],dist_min[100000],a[100000][4],
           qfac[100000],bfac[100000];
  FILE     *fp, *fp1;
  char     fn[20],nullstr[5],atomtype[100000][5],residue[100000][4],chain[100000],
           specifier[100000];

  xmin=200;
  xmax=-200;
  ymin=200;
  ymax=-200;
  zmin=200;
  zmax=-200;
  rpara=5;
  parapx=(float *)malloc(20000000*sizeof(float));
  parapy=(float *)malloc(20000000*sizeof(float));
  parapz=(float *)malloc(20000000*sizeof(float));

/*************** opening the pdb file and reading data *******************/

  printf("\n/********** Structure input *************************************/\n\n");

  printf("Enter filename+extension (in pdb format): ");
  gets(fn);
  if (strlen(fn)==0)
     puts("No filename entered.\n");
  fp=fopen(fn,"r");
  if (!fp)
     puts("Error: file name not found.\n");
  
  printf("\n/****************************************************************/\n\n");

  natoms=0;
  na=0;
  fscanf (fp," %s",&nullstr);

  while (nullstr[0]!='E'){
     if (nullstr[0]=='A') {
        natoms+=1;
        na+=1;

        fscanf (fp," %d",&atomnum[na]);
        fscanf (fp," %s",&atomtype[na]);      
           if (atomtype[na][0] == 'H' || atomtype[na][1] == 'H') 
           {
              a[na][0] =1.0;
              specifier[na]='H';
           }
           else if (atomtype[na][0] == 'C') 
           {
              a[na][0] =1.7;
              specifier[na]='C';
           }
           else if (atomtype[na][0] == 'N') 
           {
              a[na][0] =1.5;
              specifier[na]='N';
           }
           else if (atomtype[na][0] == 'O') 
           {
              a[na][0] =1.4;
              specifier[na]='O';
           }
           else if (atomtype[na][0] == 'S') 
           {
              a[na][0] =1.8;
              specifier[na]='S';
           }
           else if (atomtype[na][0] == 'P') 
           {
              a[na][0] =1.8;
              specifier[na]='P';
           }
           switch (specifier[na]) 
           {
              case 'H': a[na][0] =1.0; break;
              case 'C': a[na][0] =1.7; break;
              case 'N': a[na][0] =1.5; break;
              case 'O': a[na][0] =1.4; break;
              case 'S': a[na][0] =1.8; break;
              case 'P': a[na][0] =1.8; break;
           }
        fscanf (fp," %s",&residue[na]);
        fscanf (fp," %d",&resnum[na]);
        fscanf (fp," %f",&a[na][1]);
           if (a[na][1]<xmin) xmin=a[na][1];
           if (a[na][1]>xmax) xmax=a[na][1];
        fscanf (fp," %f",&a[na][2]);
           if (a[na][2]<ymin) ymin=a[na][2];
           if (a[na][2]>ymax) ymax=a[na][2];
        fscanf (fp," %f",&a[na][3]);
           if (a[na][3]<zmin) zmin=a[na][3];
           if (a[na][3]>zmax) zmax=a[na][3];
        fscanf (fp," %f",&qfac[na]);
        fscanf (fp," %f",&bfac[na]);
        fscanf (fp," %s",&chain[na]);        
     }
     fscanf (fp," %s",&nullstr);
  }
  fclose(fp);

/*************** Pseudo atom placement ***********************************/

  printf ("xmin is : %7.3lf\n",xmin);
  printf ("xmax is : %7.3lf\n",xmax);
  printf ("ymin is : %7.3lf\n",ymin);
  printf ("ymax is : %7.3lf\n",ymax);
  printf ("zmin is : %7.3lf\n",zmin);
  printf ("zmax is : %7.3lf\n",zmax);

  rpara = 3.5;
  layer = 20.0;
  spread = 0.5;
  radius = 50.0;
  flag_change = 0;

  printf("\n/********** Input parameters for pseudo-atom placement **********/\n\n");
  printf("  Radius of paramagnetic relaxation enhancer [angs]: %.2f\n", rpara);
  printf("  Layer surrounding the molecule      [angs]:        %.2f\n", layer);
  printf("  Increment for pseudo atom placement [angs]:        %.2f\n", spread);
  printf("  Radius for integration              [angs]:        %.2f\n\n", radius);
  printf("/****************************************************************/\n\n");
  printf("Change any parameters [Yes(1), No(0)] ? ");
  scanf("%d", &flag_change);
  printf("\n");

  if (flag_change == 1) {
     printf("\n  Radius of paramagnetic relaxation enhancer [angs]: ");
     scanf("%f", &rpara);
     printf("\n  Layer surrounding the molecule [angs]: ");
     scanf("%f", &layer);
     printf("\n  Increment for pseudo atom placement [angs]: ");
     scanf("%f", &spread);
     printf("\n  Radius for integration              [angs]: ");
     scanf("%f", &radius);
  }

  c = 1;
  z=(floor(zmin)-layer);
  while (z<=(zmax+layer)){
     y=(floor(ymin)-layer);
     printf ("z is : %.2f\n",z);

     while (y<=(ymax+layer)){
        x=(floor(xmin)-layer);

        while (x<=(xmax+layer)){
           i=1;
           posit=1;
   
           while (i<=natoms){
              d=sqrt(pow((x-a[i][1]),2)+ pow((y-a[i][2]),2) + pow((z-a[i][3]),2));
              if(d<(rpara+a[i][0])){
                 posit=0;
                 break;
              }
              i+=1;
           }
           if (posit == 1){
              parapx[c]=x;
              parapy[c]=y;
              parapz[c]=z;
              c+=1;
           }
           x+=spread;
        }
        y+=spread;
     }
     z+=spread;
  }

  printf ("parapxyz is finished.\n\n");
  printf ("number of pseudo atoms: %d\n",c);

/*************** Calculation of the influence of Paramagnet **************/
/* This version calculates the PRE for hydrogen atoms                    */
/* Change f_c in order to get PREs for 13C spins                         */

  f=2;
  
  for (j = 1; j <= natoms; j++)
  {
     paravol[j]=0;
     if (specifier[j]=='H')
     {
        dist_min[j] = 100000;
        for (k = 1; k < c; k++)
        {
           if (parapx[k]>(a[j][1]-radius) && parapx[k]<(a[j][1]+radius) && parapy[k]>(a[j][2]-radius) && parapy[k]<(a[j][2]+radius) && 
              parapz[k]>(a[j][3]-radius) && parapz[k]<(a[j][3]+radius)){
              parainfl[j]=sqrt(pow((a[j][1]-parapx[k]),2)+pow((a[j][2]-parapy[k]),2)+pow((a[j][3]-parapz[k]),2));
              if (parainfl[j] < dist_min[j]){
                 dist_min[j] = parainfl[j];
              }
              paravol[j] += pow(f/parainfl[j],6);
           }
        }
     }
  }

/*************** Normalize the PRE to the maximal measured PRE ***********/
 
  PREmaxcalc=0.0;
  
  for (q=1;q<=natoms;q++){
     if (paravol[q]>PREmaxcalc){PREmaxcalc=paravol[q];}    
  }

  printf("\nPREmax: %f\n",PREmaxcalc);

  PREmaxmeas = 15.0;
  norm       = PREmaxcalc/PREmaxmeas;
    
/*************** Write output files **************************************/
/* PREs are written out as B-factors                                     */
/* The distance to the closest paramagnet is writen out to PRE_dist.out  */
  
  fp = fopen("paravorelax.pdb","w");
  fp1= fopen("PRE_dist.out","w");

  for (q=1;q<=natoms;q++){
     fprintf (fp,"%-6s","ATOM");
     fprintf (fp," %4d",atomnum[q]);
     fprintf (fp," %4s",atomtype[q]);
     fprintf (fp," %3s",residue[q]);
     fprintf (fp," %1c",chain[q]);
     fprintf (fp," %3d",resnum[q]);
     fprintf (fp," %11.3f",a[q][1]);
     fprintf (fp," %7.3f",a[q][2]);
     fprintf (fp," %7.3f",a[q][3]);
     fprintf (fp," %5.2f",qfac[q]);
     fprintf (fp," %5.2f",paravol[q]/norm);
     fprintf (fp," %11c",specifier[q]);
     fprintf (fp," \n");
     
     fprintf (fp1,"%-6s","ATOM");
     fprintf (fp1," %4d",atomnum[q]);
     fprintf (fp1," %4s",atomtype[q]);
     fprintf (fp1," %3s",residue[q]);
     fprintf (fp1," %1c",chain[q]);
     fprintf (fp1," %3d",resnum[q]);
     fprintf (fp1," %11.3f",a[q][1]);
     fprintf (fp1," %7.3f",a[q][2]);
     fprintf (fp1," %7.3f",a[q][3]);
     fprintf (fp1," %5.2f",qfac[q]);
     fprintf (fp1," %5.2f",paravol[q]/norm);
     fprintf (fp1," %5.2f",dist_min[q]);
     fprintf (fp1," %11c",specifier[q]);
     fprintf (fp1," \n");
  }

  fprintf (fp,"%-3s","END");
  fprintf (fp1,"%-3s","END");

  fclose(fp);
  fclose(fp1);

  return 0;

}
