/*************************************************************************/
/* Reads a pdb file in the cns pdb format                                */
/* Reads a PERE restraint file of the format:                            */
/*               "PERE" Atomnumber Atomtype PRE_value(lower boundary)    */
/*               e.g. PERE 100 HN 10.3                                   */
/* Calculates the distance between the atom for which a PERE constraint  */
/* is given and the pseudo molecules. All pseudo molecules which are     */
/* closer than the lower boundary and therefore violate the PERE constr. */
/* are output in a separate file "PERE_violations.out".                  */
/*                                                                       */
/* Programmer:   Tobias Madl                                             */
/* Date:         2007                                                    */
/* Organization:           Institute of Chemistry                        */
/*                         Organic and Bioorganic Chemistry              */
/*                         Structural and Analytical Organic Chemistry   */
/*                         Heinrichstrasse 28                            */
/*                         A-8010 Graz                                   */
/*                         Austria                                       */
/*                         Europe                                        */
/*                                                                       */
/*************************************************************************/

#include <stdio.h>
#include <stddef.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>
#include <time.h>


int main() {

  time_t   start, end;
  int      na, i, j, k, p, q, natoms, c, resnum_PRE_start, temp_an,
           pseudo_first, pseudo_last, shift,
           C_PRE_numconstr, C_natoms, C_viol,C_struct,
           F_viol;
  int      atomnum[100000], resnum[100000], PRE_atomnum[100000], 
           PRE_resnum[100000];
  double   dif, disttopseu;
  float    upper,
           a[100000][4], qfac[100000], bfac[100000], paravol[100000],
           PRE_value[100000];
  FILE     *fp, *fp_PRE_viol, *fp_PRE_viol_cns;
  char     qt,molsegid,pmolsegid,
           fn[20],nullstr[100],atomtype[100000][5],residue[100000][4],
           chain[100000], specifier[100000], PRE_atomtype[100000][5];

/*************** opening the pdb file and reading data *******************/


  printf("\n\n/****** FILE INPUT PDB FILE ********************/\n\n");
  printf("Enter filename (XY.pdb): ");
  gets(fn);
  if (strlen(fn)==0)
     puts("No filename entered.\n");
  fp=fopen(fn,"r");
  if (!fp)
     puts("Error: file name not found.\n");

  natoms=0;
  na=0;
  C_struct=0;
    
  time (&start);
  
  fscanf (fp," %s",&nullstr);

  while (nullstr[0]!='E')
  {
     if (nullstr[0]=='A')  
     {
        natoms+=1;
        na+=1;

        fscanf (fp," %d",&atomnum[na]);
        if (atomnum[na]==1)
        {
           C_struct+=1;
        }
        fscanf (fp," %s",&atomtype[na]);
        fscanf (fp," %s",&residue[na]);        
        fscanf (fp," %d",&resnum[na]);
        fscanf (fp," %f",&a[na][1]);           
        fscanf (fp," %f",&a[na][2]);
        fscanf (fp," %f",&a[na][3]);
        fscanf (fp," %f",&qfac[na]);
        fscanf (fp," %f",&bfac[na]);
        fscanf (fp," %s",&chain[na]);
     }
     fscanf (fp," %s",&nullstr);
  }
  fclose(fp);
  
  C_natoms=natoms;

/*************** open the PRE constraints file and read data *************/

  printf("\n/****** FILE INPUT PRE CONSTRAINTS *************/\n\n");
  printf("Enter filename (XY.txt): ");
  gets(fn);
  if (strlen(fn)==0)
     puts("No filename entered.\n");
  fp=fopen(fn,"r");
  if (!fp)
     puts("Error: file name not found.\n");
  printf("\n/***********************************************/\n");

  C_PRE_numconstr=0;
  na=0;

  fscanf (fp," %s",&nullstr);

  while (nullstr[0]!='E'){
     if (nullstr[0]=='P') {
        na++;
        C_PRE_numconstr++; 
        fscanf (fp," %d",&PRE_atomnum[na]);
        fscanf (fp," %d",&PRE_resnum[na]);
        fscanf (fp," %s",&PRE_atomtype[na]);
        fscanf (fp," %f",&PRE_value[na]);
     }
     fscanf (fp," %s",&nullstr);
  }
  fclose(fp);

/*************** Calculate the distance to pseudo molecules **************/

   
  printf("\n\n/****** PSEUDO MOLECULES ***********************/\n\n");

  printf("Enter the molecule segid (e.g. A, 1 char): ");
  scanf("%c", &molsegid);
  printf("Enter the pseudomolecule segid (e.g. B, 1 char): ");
  scanf(" %c", &pmolsegid);
  printf("Enter the atom number of first pseudo molecule (O): ");
  scanf("%d", &pseudo_first);
  printf("Enter the atom number of last pseudo molecule (O): ");
  scanf("%d", &pseudo_last);
  printf("Enter the upper distance limit [A]: ");
  scanf("%f", &upper);
  printf("\n\n/***********************************************/\n\n");

  
  fp_PRE_viol=fopen("PERE_violations.out","w");
  fp_PRE_viol_cns=fopen("PERE_violations.cns","w");

  qt='"';

  for (k=0;k<C_struct;k++)
  {   
     shift=k*natoms/C_struct;

     for (i = 1; i <= C_PRE_numconstr; i++)
     {          
        fprintf(fp_PRE_viol, "PRE %3d %3d %4s ", PRE_atomnum[i], PRE_resnum[i], PRE_atomtype[i]);
        temp_an = PRE_atomnum[i]+shift;
        F_viol=0;

        for (j = pseudo_first+shift; j <= pseudo_last+shift; j += 3) 
        {    
           disttopseu = sqrt ( pow( a[j][1]-a[temp_an][1],2) + pow( a[j][2]-a[temp_an][2],2) +  pow( a[j][3]-a[temp_an][3],2) );
           if (disttopseu < PRE_value[i]) 
           {
              F_viol = 1;
              fprintf(fp_PRE_viol, " %d ", resnum[j]);
              fprintf(fp_PRE_viol_cns, "assign ( segid %c%c%c and resid %3d and name %4s )",qt,molsegid,qt,PRE_resnum[i],PRE_atomtype[i]);
              fprintf(fp_PRE_viol_cns, " ( segid %c%c%c and name O and resid %d ",qt,pmolsegid,qt,resnum[j]);
              fprintf(fp_PRE_viol_cns, ") %5.3f  0.000 %5.3f\n", PRE_value[i], upper);     
           }
        }
        fprintf(fp_PRE_viol, "\n");
     }
  }
 
  fprintf(fp_PRE_viol, "END\n");
  fprintf(fp_PRE_viol_cns, "END\n");


  fclose(fp_PRE_viol);
  fclose(fp_PRE_viol_cns);
 
/* Statistics ************************************************************/

  time (&end);
  dif = difftime (end,start);
  
  printf("\n/******** Statistics ****************************/\n");
  printf("\n    Number of structures (pdb): %d", C_struct);  
  printf("\n    Number of atoms (pdb):      %d", natoms);  
  printf("\n    Calculation time [s]:       %.2lf\n", dif);
  printf("\n/************************************************/\n");
  
  printf("\n/****************** Finished ********************/\n");
  printf("\n");
    
  return 0;

}
