/*************************************************************************/
/* Reads a pdb file in the cns pdb format                                */
/* Calculates the boundaries in x, y, z (xmin, xmax, ymin, ymax, ...).   */
/* Calculates the corrdinates of pseudo atoms with a specific radius     */
/* placed on the grid spanned by the above boundaries.                   */
/*                                                                       */
/* Programmer:   Tobias Madl and Klaus Zangger                           */
/* Date:         2008                                                    */
/* Organization:           Institute of Chemistry                        */
/*                         Organic and Bioorganic Chemistry              */
/*                         Structural and Analytical Organic Chemistry   */
/*                         Heinrichstrasse 28                            */
/*                         A-8010 Graz                                   */
/*                         Austria                                       */
/*                         Europe                                        */
/*                                                                       */ 
/*************************************************************************/

#include <stdio.h>
#include <stddef.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>
#include <time.h>
#define LAENGE 80
#define breakone "-1 ' ' ' ' ' ' ' ' ' ' -1.00000 -1.00000\n"
#define breaktwo "-1 -1\n"
#define breakthree "-1 -1 -1\n"
#define breakpdb "END\n"


int main() {

  unsigned _stklen = 4194304;                        /* need a 1MB stack */
  time_t   start, end;
  double   dif;
  int      atomnum[20000],na,i,j,k,p,q,natoms,c,resnum_PRE_start,
           test_same,resnum[20000],atfirst,atlast,
           F_posit,
           C_natoms, C_pseudoat, C_PRE_out, C_PRE_resnum, C_noe;
  float    a[20000][4], qfac[20000], bfac[20000],
           coord_pseudoat[20000][4],
           xmin, ymin, zmin, xmax, ymax, zmax, incr_xmin, incr_ymin, 
           incr_zmin, layer, radius, radius_vdw, gridspace, h_x, h1_y, 
           h2_y, dist_x, dist_y, dist, dist_pseu, PRE_qfac, PRE_paravol,
           dist_pseudo, T_dist_noe, dist_to_PRE, dist_to_PRE_min,
           dist_to_PRE_max, dist_temp,  
           cnst, paravol[20000], dist_molpseu, temp;
  FILE     *fp, *fp_pseudocoord, *fp_pseudomtf, *fp_pseudomtf_1, *fp_pseudomtf_2, 
           *fp_pseudomtf_3,*fp_mtf_out;
  char     PRE_chain, quotation,
           fn[20],fnmtf[20],nullstr[10],atomtype[20000][5],residue[20000][4],
           chain[20000], specifier[20000], in_PRE_chain[5],text[LAENGE],
           mtf_1[LAENGE],mtf_2[LAENGE],mtf_3[LAENGE];

/*************** open the pdb file and read data **************************/


  printf("\n\n/****** FILE INPUT *****************************/\n\n");
  printf("Enter filename (XY.pdb): ");
  gets(fn);
  if (strlen(fn)==0)
     puts("No filename entered.\n");
  fp=fopen(fn,"r");
  if (!fp)
     puts("Error: file name not found.\n");
  printf("Enter filename (XY.mtf): ");
  gets(fnmtf);
  if (strlen(fnmtf)==0)
     puts("No filename entered.\n");
  fp_pseudomtf=fopen(fnmtf,"r");
  if (!fp_pseudomtf)
     puts("Error: file name not found.\n");
  fclose(fp_pseudomtf);   
  printf("\n\n/***********************************************/\n\n");

  natoms=0;
  na=0;
    
  time (&start);
  
  fscanf (fp," %s",&nullstr);

  while (nullstr[0]!='E'){
     if (nullstr[0]=='A') {
        natoms+=1;
        na+=1;

        fscanf (fp," %d",&atomnum[na]);
        fscanf (fp," %s",&atomtype[na]);
        fscanf (fp," %s",&residue[na]);        
        fscanf (fp," %d",&resnum[na]);
        fscanf (fp," %f",&a[na][1]);           
        fscanf (fp," %f",&a[na][2]);
        fscanf (fp," %f",&a[na][3]);
        fscanf (fp," %f",&qfac[na]);
        fscanf (fp," %f",&bfac[na]);
        fscanf (fp," %s",&chain[na]);
     }
     fscanf (fp," %s",&nullstr);
  }
  fclose(fp);
  
  C_natoms=natoms;
  
  printf("/**** NATOMS ********/\n");
  printf(" natoms: %d\n", natoms);
  printf("/********************/\n\n");
   
/* Calc. coord. of pseudo atoms placed on the grid spanned by boundaries */

  layer=10.0;
  gridspace=3.5;
  radius=3.5;
  dist_to_PRE=3.5;
  
  printf("\n\n/****** INPUT FOR PSEUDO ATOM POSITIONING ******/\n\n");
  printf("  Layer around protein [A]:  ");
  scanf("%f", &layer);
  printf("  Gridspace [A]:  ");
  scanf("%f", &gridspace);
  printf("  Radius of pseudo atoms [A]:  ");
  scanf("%f", &radius);
  printf("  Atom number of first residue:  ");
  scanf("%d", &atfirst);
  printf("  Atom number of last residue:  ");
  scanf("%d", &atlast); 
  printf("\n/************************************************/\n");  

  xmax=a[atfirst][1];
  xmin=a[atfirst][1];
  ymax=a[atfirst][2];
  ymin=a[atfirst][2];
  zmax=a[atfirst][3];
  zmin=a[atfirst][3];
    
  for (q=atfirst+1; q<=atlast; q++) {                     /* Calculate boundaries*/
     if (a[q][1] > xmax) {xmax = a[q][1];}
     if (a[q][1] < xmin) {xmin = a[q][1];}
     if (a[q][2] > ymax) {ymax = a[q][2];}
     if (a[q][2] < ymin) {ymin = a[q][2];}
     if (a[q][3] > zmax) {zmax = a[q][3];}
     if (a[q][3] < zmin) {zmin = a[q][3];}   
  }
  
  printf("/**** Boundaries ****/\n");
  printf(" xmin: %.2f\n", xmin);
  printf(" xmax: %.2f\n", xmax);
  printf(" ymin: %.2f\n", ymin);
  printf(" ymax: %.2f\n", ymax);
  printf(" zmin: %.2f\n", zmin);
  printf(" zmax: %.2f\n", zmax);
  printf("/********************/\n\n");

  printf("\nrunning! Please wait. This may take a few minutes...\n\n");

  radius_vdw=1.2;
  incr_xmin=xmin-layer;                    /* +10A to ensure that pseudo */
  incr_ymin=ymin-layer;                    /* atoms do not overlap       */     
  incr_zmin=zmin-layer;
  
  coord_pseudoat[1][1]=incr_xmin;             /* Place first pseudo atom */
  coord_pseudoat[1][2]=incr_ymin;
  coord_pseudoat[1][3]=incr_zmin;
  C_pseudoat=1;

  dist_to_PRE_min=radius_vdw;
  dist_to_PRE_max=radius_vdw+layer;
      
  while (incr_zmin <= zmax + layer) {
     while (incr_ymin <= ymax + layer) {
        while (incr_xmin <= xmax + layer) {
           F_posit=1;
           dist_temp=1000.0;

           for (i=1; i<=C_natoms; i++){
              dist=sqrt(pow((incr_xmin-a[i][1]),2)+pow((incr_ymin-a[i][2]),2)+pow((incr_zmin-a[i][3]),2));
              if (dist<dist_temp) { dist_temp=dist;} 
              if (dist<dist_to_PRE_min) {
                 incr_xmin+=gridspace;
                 F_posit=0;
                 break;
              }
           }

           if (dist_temp > dist_to_PRE_max) {
              F_posit=0;
              incr_xmin+=gridspace;
           }           

           if (F_posit!=0) {
              for (j=1; j<=C_pseudoat; j++) {
                 dist_pseu=sqrt(pow((coord_pseudoat[j][1]-incr_xmin),2)+pow((coord_pseudoat[j][2]-incr_ymin),2)+pow((coord_pseudoat[j][3]-incr_zmin),2));   
                 if (dist_pseu<radius*2) {
                    incr_xmin+=gridspace;
                    F_posit=0;
                    break;
                 }
              }
           } /* end if */
           
           if (F_posit==1) {                  /* If no atom within range */
              C_pseudoat+=1;                  /* place pseudo atom.      */
              coord_pseudoat[C_pseudoat][1]=incr_xmin;   
              coord_pseudoat[C_pseudoat][2]=incr_ymin;
              coord_pseudoat[C_pseudoat][3]=incr_zmin;
              incr_xmin+=gridspace;
           }           
        }                                                    /* end while*/
        incr_ymin+=gridspace;
        incr_xmin=xmin-layer;
     }                                                      /* end while */
     incr_zmin+=gridspace;
     incr_ymin=ymin-layer;
     incr_xmin=xmin-layer;     
  }                                                         /* end while */

      
/* Statistics ************************************************************/

  time (&end);
  dif = difftime (end,start);
  
  printf("\n/******** Statistics ****************************/\n");
  printf("\n    Number of atoms (pdb):   %d", natoms);  
  printf("\n    Number of pseudo atoms:  %d", C_pseudoat);
  printf("\n    Calculation time [s]:    %.2lf\n", dif);
  printf("\n/************************************************/\n");
  
/* Generate pdb and mtf files for pseudo atoms ***************************/  
/* Write NOE file with NOEs between pseudo atoms in order to retain ******/  
/* compactness of the pseudo-atom box ************************************/

  PRE_qfac=0.0;
  PRE_paravol=0.0;
  
  printf("\n\n/************** INPUT ***************************/\n\n");
  printf("  Start numbering of pseudo residues at:  ");
  scanf("%d", &resnum_PRE_start);
  printf("  Chain ID for pseudo atoms (1 char):  ");
  scanf("%s", in_PRE_chain);
  printf("\n/************************************************/\n");  
  
  PRE_chain=in_PRE_chain[0];  
  dist_x=0.5858823;                          /* Dist(HO)*cos(ANG(HOH/2)) */
  dist_y=0.75695;                            /* Dist(HO)*sin(ANG(HOH/2)) */                                           
  C_PRE_resnum=resnum_PRE_start;                     
  C_pseudoat=C_pseudoat*3;                  /* Need 3 atoms for molecule */
  quotation='"';
  p=0;
  
  /* opening original pdb file and write it into pseudo pdb file start */
  
  fp=fopen(fn,"r");
  fp_pseudocoord=fopen("PERE.pdb","w");
  
  while (fgets(text,LAENGE,fp)) {
    if (strcmp(breakpdb,text)!=0)
       fputs(text,fp_pseudocoord);
  }

  /* opening original pdb file and write it into pseudo pdb file end */
  
  fp_pseudomtf_1=fopen("PERE_pseudo_main.mtf","w");
  fp_pseudomtf_2=fopen("PERE_pseudo_bond.mtf","w");
  fp_pseudomtf_3=fopen("PERE_pseudo_angle.mtf","w");
  
  for (q=1; q<=C_pseudoat; q+=3){   
     p+=1;
     C_PRE_out=q+natoms;                          /* PDB: Coordinate O */     
     fprintf (fp_pseudocoord,"%-6s", "ATOM");
     fprintf (fp_pseudocoord," %4d", C_PRE_out);
     fprintf (fp_pseudocoord," %-4s", " O");
     fprintf (fp_pseudocoord," %4s", "HOH  ");
/*     fprintf (fp_pseudocoord," %1c", PRE_chain); */
     fprintf (fp_pseudocoord," %3d", C_PRE_resnum); 
     fprintf (fp_pseudocoord," %11.3f", coord_pseudoat[p][1]);
     fprintf (fp_pseudocoord," %7.3f", coord_pseudoat[p][2]);
     fprintf (fp_pseudocoord," %7.3f", coord_pseudoat[p][3]);
     fprintf (fp_pseudocoord," %5.2f", PRE_qfac);
     fprintf (fp_pseudocoord," %5.2f", PRE_paravol);
     fprintf (fp_pseudocoord," %6c", PRE_chain);
     fprintf (fp_pseudocoord," \n");
     h_x=coord_pseudoat[p][1]+dist_x;
     h1_y=coord_pseudoat[p][2]+dist_y;
     C_PRE_out+=1;
     fprintf (fp_pseudocoord,"%-6s", "ATOM");      /* PDB: Coordinate H1 */
     fprintf (fp_pseudocoord," %4d", C_PRE_out);
     fprintf (fp_pseudocoord," %-4s", " H1");
     fprintf (fp_pseudocoord," %3s", "HOH  ");
/*    fprintf (fp_pseudocoord," %1c", PRE_chain); */
     fprintf (fp_pseudocoord," %3d", C_PRE_resnum);
     fprintf (fp_pseudocoord," %11.3f", h_x);
     fprintf (fp_pseudocoord," %7.3f", h1_y);
     fprintf (fp_pseudocoord," %7.3f", coord_pseudoat[p][3]);
     fprintf (fp_pseudocoord," %5.2f", PRE_qfac);
     fprintf (fp_pseudocoord," %5.2f", PRE_paravol);
     fprintf (fp_pseudocoord," %6c", PRE_chain);
     fprintf (fp_pseudocoord," \n");
     h2_y=coord_pseudoat[p][2]-dist_y;
     C_PRE_out+=1;
     fprintf (fp_pseudocoord,"%-6s", "ATOM");      /* PDB: Coordinate H2 */
     fprintf (fp_pseudocoord," %4d", C_PRE_out);
     fprintf (fp_pseudocoord," %-4s", " H2");
     fprintf (fp_pseudocoord," %3s", "HOH  ");
/*    fprintf (fp_pseudocoord," %1c", PRE_chain); */
     fprintf (fp_pseudocoord," %3d", C_PRE_resnum);
     fprintf (fp_pseudocoord," %11.3f", h_x);
     fprintf (fp_pseudocoord," %7.3f", h2_y);
     fprintf (fp_pseudocoord," %7.3f", coord_pseudoat[p][3]);
     fprintf (fp_pseudocoord," %5.2f", PRE_qfac);
     fprintf (fp_pseudocoord," %5.2f", PRE_paravol);
     fprintf (fp_pseudocoord," %6c", PRE_chain);
     fprintf (fp_pseudocoord," \n");
      
     /* MTF: main */     
     fprintf(fp_pseudomtf_1, "%d '%1c' '%d' 'HOH' 'O' 'OT' -.834000 15.9994\n", C_PRE_out-2, PRE_chain, C_PRE_resnum);
     fprintf(fp_pseudomtf_1, "%d '%1c' '%d' 'HOH' 'H1' 'HT' .417000 1.00800\n", C_PRE_out-1, PRE_chain, C_PRE_resnum);
     fprintf(fp_pseudomtf_1, "%d '%1c' '%d' 'HOH' 'H2' 'HT' .417000 1.00800\n", C_PRE_out, PRE_chain, C_PRE_resnum);

     /* MTF: bonds*/
     fprintf(fp_pseudomtf_2, "%d %d\n", C_PRE_out-2, C_PRE_out-1);
     fprintf(fp_pseudomtf_2, "%d %d\n", C_PRE_out-2, C_PRE_out);

     /* MTF: angles */
     fprintf(fp_pseudomtf_3, "%d %d %d\n", C_PRE_out-1, C_PRE_out-2, C_PRE_out);     
  
          
     C_PRE_resnum+=1;     
  }

  fprintf (fp_pseudocoord,"%-3s","END");
  


  fclose(fp_pseudocoord);
  fclose(fp_pseudomtf_1);
  fclose(fp_pseudomtf_2);
  fclose(fp_pseudomtf_3);
  
  /* combining individual mtf files start */
  
  fp_pseudomtf=fopen(fnmtf,"r");
  fp_mtf_out=fopen("PERE.mtf","w");

  while (strcmp(breakone,text)!=0) {
    fgets(text,LAENGE,fp_pseudomtf);
    if (strcmp(breakone,text)!=0)
       fputs(text,fp_mtf_out);
  }
  
  fp_pseudomtf_1=fopen("PERE_pseudo_main.mtf","r");
  while (fgets(mtf_1,LAENGE,fp_pseudomtf_1))
        fputs(mtf_1,fp_mtf_out);
  fclose(fp_pseudomtf_1);
  fputs(breakone,fp_mtf_out);
        
  while (strcmp(breaktwo,text)!=0) {
    fgets(text,LAENGE,fp_pseudomtf);
    if (strcmp(breaktwo,text)!=0)
       fputs(text,fp_mtf_out);
  }
  
  fp_pseudomtf_2=fopen("PERE_pseudo_bond.mtf","r");
  while (fgets(mtf_2,LAENGE,fp_pseudomtf_2))
        fputs(mtf_2,fp_mtf_out);
  fclose(fp_pseudomtf_2);
  fputs(breaktwo,fp_mtf_out);
  
  while (strcmp(breakthree,text)!=0) {
    fgets(text,LAENGE,fp_pseudomtf);
    if (strcmp(breakthree,text)!=0)
       fputs(text,fp_mtf_out);
  }
  
  fp_pseudomtf_3=fopen("PERE_pseudo_angle.mtf","r");
  while (fgets(mtf_3,LAENGE,fp_pseudomtf_3))
        fputs(mtf_3,fp_mtf_out);
  fclose(fp_pseudomtf_3);
  fputs(breakthree,fp_mtf_out);
  
  while (fgets(text,LAENGE,fp_pseudomtf))
        fputs(text,fp_mtf_out);

  fclose(fp_pseudomtf);
  fclose(fp_mtf_out);
  
  /* combining individual mtf files end */
  
  printf("\n/****************** Finished ********************/\n");
  printf("\n");    
  return 0;

}
